from aspen.resources.dynamic_resource import DynamicResource


class SocketResource(DynamicResource):

    max_pages = 1
    max_pages = 3

    def respond(self):
        """Override and kill it. For sockets the Socket object responds.
        """
        raise NotImplemented

    def compile_third(self, one, two, three, padding):
        """Given three bytestrings, return a code object.

        This method depends on self.fs.

        """
        # See DyanmicResource._compile for comments on this algorithm.
        three = three.replace('\r\n', '\n')
        three = padding + three
        three = compile(three, self.fs, 'exec')
        return three

    def exec_second(self, socket, request):
        """Given a Request, return a namespace dictionary.
        """
        namespace = self.one.copy()
        namespace.update(request.namespace)
        namespace['request'] = request 
        namespace['socket'] = socket
        namespace['channel'] = socket.channel
        exec self.two in namespace
        return namespace




import logging
import optparse
import os
import socket
import sys

import aspen
from aspen.configuration.exceptions import ConfigurationError


LOG_LEVELS = ( 'NIRVANA'    # oo
             , 'CRITICAL'   # 50
             , 'ERROR'      # 40
             , 'WARNING'    # 30
             , 'INFO'       # 20
             , 'DEBUG'      # 10
             , 'NOTSET'     #  0
              )


def validate_address(address):
    """Given a socket address string, return a tuple (sockfam, address).

    This is called from a couple places, and is a bit complex.

    """

    if address[0] in ('/','.'):
        if aspen.WINDOWS:
            raise ConfigurationError("Can't use an AF_UNIX socket on Windows.")
            # but what about named pipes?
        sockfam = socket.AF_UNIX
        # We could test to see if the path exists or is creatable, etc.
        address = os.path.realpath(address)

    elif address.count(':') > 1:
        sockfam = socket.AF_INET6
        # @@: validate this, eh?

    else:
        sockfam = socket.AF_INET
        # Here we need a tuple: (str, int). The string must be a valid
        # IPv4 address or the empty string, and the int -- the port --
        # must be between 0 and 65535, inclusive.


        err = ConfigurationError("Bad address %s" % str(address))


        # Break out IP and port.
        # ======================

        if isinstance(address, (tuple, list)):
            if len(address) != 2:
                raise err
            ip, port = address
        elif isinstance(address, basestring):
            if address.count(':') != 1:
                raise err
            ip_port = address.split(':')
            ip, port = [i.strip() for i in ip_port]
        else:
            raise err


        # IP
        # ==

        if not isinstance(ip, basestring):
            raise err
        elif ip == '':
            ip = '0.0.0.0' # IP defaults to INADDR_ANY for AF_INET; specified
                           # explicitly to avoid accidentally binding to
                           # INADDR_ANY for AF_INET6.
        else:
            try:
                socket.inet_aton(ip)
            except socket.error:
                raise err


        # port
        # ====
        # Coerce to int. Must be between 0 and 65535, inclusive.

        if isinstance(port, basestring):
            if not port.isdigit():
                raise err
            else:
                port = int(port)
        elif isinstance(port, int) and not (port is False):
            # already an int for some reason (called interactively?)
            pass
        else:
            raise err

        if not(0 <= port <= 65535):
            raise err


        # Success!
        # ========

        address = (ip, port)


    return address, sockfam


def validate_log_level(log_level):
    """Convert a string to an int.
    """
    if log_level not in LOG_LEVELS:
        msg = "logging level must be one of %s, not %s"
        log_levels = ', '.join(LOG_LEVELS)
        raise ConfigurationError(msg % (log_levels, log_level))
    if log_level == 'NIRVANA':
        log_level = sys.maxint
    else:
        log_level = getattr(logging, log_level)
    return log_level


# optparse
# ========

def callback_address(option, opt, value, parser_):
    """Must be a valid AF_INET or AF_UNIX address.
    """
    address, sockfam = validate_address(value)
    parser_.values.address = address
    parser_.values.sockfam = sockfam
    parser_.values.have_address = True
    parser_.values.raw_address = value 

def callback_log_level(option, opt, value, parser_):
    """Convert the string to an int.
    """
    parser_.values.log_level = validate_log_level(value)
    parser_.values.raw_log_level = value

def store_raw(option, opt, value, parser_):
    """Store both the computed and the raw value for get_clean_argv in __init__.
    """
    setattr(parser_.values, option.dest, value)
    setattr(parser_.values, 'raw_'+option.dest, value)


usage = "aspen [options]; --help for "\
        "more"
version = """\
aspen, version %s

(c) 2006-2011 Chad Whitacre and contributors
http://aspen.io/
""" % aspen.__version__

optparser = optparse.OptionParser(usage=usage, version=version)
optparser.description = """\
Aspen is a Python web framework, and this is the server it comes with, which is
suitable for production. If given no arguments or options, it will start
serving a website from the current directory on port 8080, based on
configuration files in ./.aspen/etc/, logging to stdout. Full documentation is
on the web at http://aspen.io/.
"""


# Basic
# -----

basic_group = optparse.OptionGroup( optparser
                                  , "Basics"
                                  , "How should we get on the network?"
                                   )
basic_group.add_option( "-a", "--address"
                    , action="callback"
                    , callback=callback_address
                    , default=('0.0.0.0', 8080)
                    , dest="address"
                    , help="the IPv4 or Unix address to bind to [0.0.0.0:8080]"
                    , type='string'
                     )
basic_group.add_option( "-e", "--engine"
                    , action="callback"
                    , callback=store_raw
                    , choices=aspen.ENGINES
                    , default=None
                    , dest="engine"
                    , help=( "the HTTP engine to use, one of "
                           + "{%s}" % ','.join(aspen.ENGINES)
                           + " [%s]" % aspen.ENGINES[0]
                            )
                    , type='choice'
                     )

optparser.add_option_group(basic_group)


# Logging 
# -------

logging_group = optparse.OptionGroup( optparser
                                    , "Logging"
                                    , "Configure the Python logging library. "
                                      "For more complex needs use a "
                                      "logging.conf file."
                                     )

logging_group.add_option( "-o", "--log-file"
                    , action="callback"
                    , callback=store_raw
                    , default=None
                    , dest="log_file"
                    , help="the file to which messages will be logged; if "\
                           "specified, it will be rotated nightly for 7 days "\
                           "[stdout]"
                    , type='string'
                    , metavar="FILE"
                     )
logging_group.add_option( "-i", "--log-filter"
                    , action="callback"
                    , callback=store_raw
                    , default=None
                    , dest="log_filter"
                    , help="the subsystem outside of which messages will "\
                           "not be logged []"
                    , type='string'
                    , metavar="FILTER"
                     )
logging_group.add_option( "-t", "--log-format"
                    , action="callback"
                    , callback=store_raw
                    , default=None
                    , dest="log_format"
                    , help="the log message format per the Formatter class "\
                           "in the Python standard library's logging module "\
                           "[%(message)s]"
                    , type='string'
                    , metavar="FORMAT"
                     )
logging_group.add_option( "-v", "--log-level"
                    , action="callback"
                    , callback=callback_log_level
                    , choices=LOG_LEVELS
                    , default=None
                    , help="the importance level at or above which to log "\
                           "a message; options are %s [WARNING]" % \
                           ', '.join(LOG_LEVELS)
                    , type='choice'
                    , metavar="LEVEL"
                     )
optparser.add_option_group(logging_group)






from aspen import Response
from aspen.resources.resource import Resource


PAGE_BREAK = chr(12)

class StringDefaultingList(list):
    def __getitem__(self, key):
        try:
            return super(StringDefaultingList, self).__getitem__(key)
        except KeyError:
            return str(key)
ORDINALS = StringDefaultingList([ 'zero' , 'one' , 'two', 'three', 'four'
                                , 'five', 'six', 'seven', 'eight', 'nine'])


class DynamicResource(Resource):
    """This is a base class for json, socket, and template resources.
    """

    min_pages = None # set on subclass
    max_pages = None
    
    def __init__(self, *a, **kw):
        assert self.max_pages in (2, 3) # sanity check
        super(DynamicResource, self).__init__(*a, **kw)
        one, two, three = self._parse(self.raw)
        self.one, self.two, self.three = self._compile(one, two, three)

    def respond(self, request, response=None):
        """Given a Request and maybe a Response, return or raise a Response.
        """
        response = response or Response()

       
        # Populate namespace.
        # ===================
        
        namespace = self.one.copy()
        namespace.update(request.namespace)
        namespace['request'] = request
        namespace['response'] = response
   

        # Exec the script.
        # ================
    
        exec self.two in namespace


        # Hook.
        # =====
    
        return self.get_response(namespace)

        
    def _parse(self, raw):
        """Given a bytestring, return a list of three items.
        
        If there are too few pages, raise AssertionError. Any resource with
        only one page should land in StaticResource, not here.
        
        If there are too many pages, raise SyntaxError. 

        If there are fewer than self.max_pages, then pad the front of the list
        with empty strings.

        If self.max_pages is less than three (i.e., two), pad the end of the
        list with None.
        
        """

        # Support caret-L in addition to .
        uncareted = raw.replace("^L", PAGE_BREAK)
        pages = uncareted.split(PAGE_BREAK)
        npages = len(pages)

        # Check for too few pages.
        assert npages >= self.min_pages # sanity check; bug if False

        # Check for too many pages.
        if npages > self.max_pages:     # user error if True
            type_name = self.__class__.__name__[:-len('resource')]
            msg = "%s resources must have exactly %s pages; %s has %s."
            msg %= ( type_name
                   , ORDINALS[self.max_pages]
                   , self.fs
                   , ORDINALS[npages]
                    )
            raise SyntaxError(msg)

        # Pad the front with empty strings.
        while len(pages) < self.max_pages:
            pages.insert(0, '')

        # Pad the back with None.
        while len(pages) < 3:
            pages.append(None)

        return pages
       
    def _compile(self, one, two, three):
        """Given three items, return a 3-tuple of compiled objects.

        All dynamic resources compile the first two pages the same way. It's
        the third page that differs, so we require subclasses to provide a hook
        for that.

        """

        # Standardize newlines.
        # =====================
        # compile requires \n, and doing it now makes the next line easier.

        one = one.replace('\r\n', '\n')
        two = two.replace('\r\n', '\n')


        # Compute paddings and pad the second page.
        # =========================================
        # This is so we get accurate tracebacks. We will pass padding_two to 
        # the compile_third hook in case subclasses want to use it.
        
        padding = lambda s: ''.join(['\n' for n in range(s.count('\n'))])
        padding_two = padding(one)
        padding_three = padding_two + padding(two)
        two = padding_two + two 


        # Exec the first page and compile the second.
        # ===========================================
        # Below in render we take care not to mutate namespace.

        namespace = dict()
        namespace['__file__'] = self.fs
        namespace['website'] = self.website
        
        one = compile(one, self.fs, 'exec')
        two = compile(two, self.fs, 'exec')

        exec one in namespace
        one = namespace


        # Third
        # =====

        three = self.compile_third(one, two, three, padding_two)


        return one, two, three


    # Hooks
    # =====

    def compile_third(self, raw):
        """Given a bytestring, return an object.
        """
        raise NotImplementedError

    def get_response(self, namespace):
        """Given a namespace dictionary, return a Response object.
        """
        raise NotImplementedError



"""Template resources.

Problems with tornado.template:

    - no option to fail silently
    - Loader cache doesn't account for modtime
    - Is this a bug?

        {{ foo }}
        {% for foo in [1,2,3] %}
        {% end %}

    - no loop counters, eh? must do it manually with {% set %}
    - can't do this:

        {% if ... %}
            {% extends %}
        {% else %}
            {% extends %}
        {% end %}

"""
import copy 

from aspen import Response
from aspen.resources.dynamic_resource import DynamicResource
from aspen._tornado.template import Template


class TemplateResource(DynamicResource):
    """This is a template resource. It has one, two, or three pages.
    """

    max_pages = 2
    max_pages = 3

    def compile_third(self, one, two, three, padding):
        """Given three bytestrings, return a Template instance.

        This method depends on fs and website attributes on self.
        
        We used to take advantage of padding, but:

            a) Tornado templates have some weird error handling that we haven't
            exposed yet.
            
            b) It's counter-intuitive if your template resources show up in the
            browser with tons of whitespace at the beginning of them.

        """
        return Template( self._trim_initial_newline(three)
                       , name = self.fs
                       , loader = self.website.template_loader 
                       , compress_whitespace = False
                        )

    def _trim_initial_newline(self, template):
        """Trim any initial newline from page three.
        
        This is a convenience. It's nice to put ^L on a line by itself, but
        really you want the template to start on the next line.

        """
        try:
            if template[0] == '\r':     # Windows
                if template[1] == '\n':
                    template = template[2:]
            elif template[0] == '\n':   # Unix
                template = template[1:]
        except IndexError:              # empty template
            pass
        return template


    def get_response(self, namespace):
        """Given a namespace dict, return a response object.
        """
        response = namespace['response']
        response.body = self.three.generate(**namespace)
        if response.headers.one('Content-Type') is None:
            response.headers.set('Content-Type', self.mimetype)
        return response



import time
import uuid

from aspen import json, resources, Response
from aspen.sockets import HEARTBEAT, TIMEOUT, TRANSPORTS
from aspen.sockets.event import Event
from aspen.sockets.message import Message
from aspen.sockets.packet import Packet


class Socket(object):
    """Model a persistent Socket.IO socket (regardless of transport).

    Socket objects sit between Aspen's HTTP machinery and your Resource. They
    function as middleware, and the recv/send and _recv/_send semantics reflect
    this. They (the sockets) are persistent.
    
    """

    transports = ",".join(TRANSPORTS)
    heartbeat = str(HEARTBEAT)
    timeout = str(TIMEOUT)


    def __init__(self, request, channel):
        """Takes the handshake request and the socket's channel.
        """
        self.sid = uuid.uuid4().hex
        self.endpoint = request.path.raw
        self.resource = resources.get(request)
        request.website.copy_configuration_to(self)
        request.website.copy_configuration_to(channel)

        self.loop = request.engine.Loop(self)
        self.incoming = request.engine.Buffer('incoming', self)
        self.outgoing = request.engine.Buffer('outgoing', self)
        self.channel = channel
        self.channel.add(self)
        self.namespace = self.resource.exec_second(self, request)

    def shake_hands(self):
        """Return a handshake response.
        """
        handshake = ":".join([ self.sid
                             , self.heartbeat
                             , self.timeout 
                             , self.transports
                              ])
        return Response(200, handshake)

    def tick(self):
        """Exec the third page of the resource.

        It is expected that socket resources will block via self.recv() or some
        other mechanism, like reading a remote TCP socket.

        """
        exec self.resource.three in self.namespace

    def disconnect(self):
        self.loop.stop()
        self.channel.remove(self)


    # Client Side
    # ===========
    # Call these inside of your Resource.

    def sleep(self, seconds):
        """Block until seconds have elapsed.
        """
        self.engine.sleep(seconds)

    def recv(self):
        """Block until the next message is available, then return it.
        """
        return self.incoming.next()


    def send(self, data):
        """Buffer a plain message to be sent to the client.
        """
        self.__send(3, data)

    def send_utf8(self):
        """Buffer a UTF-8 message to be sent to the client.
        """
        self.__send(3, data.encode('utf8'))

    def send_json(self, data):
        """Buffer a JSON message to be sent to the client.
        """
        if not isinstance(data, basestring):
            data = json.dumps(data)
        self.__send(4, data)

    def send_event(self, data):
        """Buffer an event message to be sent to the client.
        """
        if not isinstance(data, basestring):
            data = json.dumps(data)
        self.__send(5, data)

    def __send(self, type_, data):
        message = Message()
        message.type = type_ 
        message.endpoint = self.endpoint
        message.data = data
        self.outgoing.put(message)


    # Server Side 
    # ===========
    # These are called by Aspen's HTTP machinery.

    def _recv(self):
        """Return an iterator of bytes or None. Don't block.
        """
        return self.outgoing.flush()

    def _send(self, bytes):
        """Given a packet bytestring, process messages.
        """
        packet = Packet(bytes)
        for message in packet:
            if message.endpoint != self.endpoint:
                msg = "The %s endpoint got a message intended for %s."
                msg %= self.endpoint, message.endpoint
                raise RuntimeError(msg)
            if message.type == 0:           # disconnect
                pass
            elif message.type == 1:         # connect 
                pass
            elif message.type == 2:         # heartbeat
                pass
            elif message.type in (3, 4, 5): # data message
                self.incoming.put(message.data)
                self.channel.incoming.put(message.data)
            elif message.type in (6, 7, 8): # blah, blah, blah
                pass



import datetime
import logging
import mimetypes
import os
import stat
import sys
import traceback
from os.path import join

PAGE_BREAK = chr(12)

from aspen import json
from aspen.resources.json_resource import JSONResource
from aspen.resources.template_resource import TemplateResource
from aspen.resources.socket_resource import SocketResource
from aspen.resources.static_resource import StaticResource


log = logging.getLogger('aspen.resources')


class LoadError(StandardError):
    """Represent a problem loading a resource.
    """


# Cache helpers
# =============

__cache__ = dict()        # cache, keyed to filesystem path

class Entry:
    """An entry in the global resource cache.
    """

    fspath = ''         # The filesystem path [string]
    modtime = None      # The timestamp of the last change [int]
    quadruple = None    # A post-processed version of the data [4-tuple]
    exc = None          # Any exception in reading or compilation [Exception]

    def __init__(self):
        self.fspath = ''
        self.modtime = 0
        self.quadruple = ()


# Core loaders
# ============

def get_resource_class(raw, mimetype):
    """Given bytes and a mimetype, return a Resource subclass.

    If the mimetype does not start with 'text/', then it is only a resource if
    it has at least one form feed in it. Binary files generally can't be
    decoded using UTF-8. If Python's mimetypes module doesn't know about a
    certain extension, then we default to a configurable value (default is
    text/plain).

    """

    is_dynamic = True

    if mimetype == 'application/x-socket.io':
        
        # *.sock files are always dynamic.

        pass

    elif mimetype.startswith('text/') or mimetype == 'application/json':

        # For text formats we can perform a highly accurate test for
        # dynamicity.

        c = lambda s: s in raw
        is_dynamic = c("") or c("^L")

    else:

        # For binary formats we must rely on a less-accurate test. This is
        # because a binary file can have s in it without being a resource--
        # and I've actually seen, in the wild, a file with exactly two s. So
        # we sniff the first few bytes.

        s = lambda s: raw.startswith(s)
        is_dynamic = s('"""') or s('import') or s('from')

    if not is_dynamic:
        Class = StaticResource
    elif mimetype == 'application/json':
        Class = JSONResource
    elif mimetype == 'application/x-socket.io':
        Class = SocketResource
    else:
        Class = TemplateResource

    return Class


def load(request, modtime):
    """Given a Request and a modtime, return a Resource object (w/o caching).

        Resource ------> DynamicResource ------> JSONResource
                 \                       \ 
                  \                       \----> SocketResource
                   \                       \
                    \--> StaticResource     \--> TemplateResource
        
    """

    # Load bytes.
    # ===========
    # We work with resources exclusively as bytestrings. Any unicode objects
    # passed in by the user as {{ expressions }} in Resources will be encoded
    # with UTF-8 by Tornado.

    raw = open(request.fs, 'rb').read()
    
   
    # Compute a mimetype.
    # ===================

    mimetype = mimetypes.guess_type(request.fs, strict=False)[0]
    if mimetype is None:
        mimetype = request.default_mimetype


    # Compute and instantiate a class.
    # ================================
    # An instantiated resource is compiled as far as we can take it.
 
    Class = get_resource_class(raw, mimetype)
    resource = Class(request.website, request.fs, raw, mimetype, modtime)
    return resource


def get(request):
    """Given a Request, return a Resource object (with caching).

    We need the request because it carries default_mimetype.

    """

    # XXX This is not thread-safe. It used to be, but then I simplified it
    # when I switched to diesel. Now that we have multiple engines, some of
    # which are threaded, we need to make this thread-safe again.
    
    # Get a cache Entry object.
    # =========================

    if request.fs not in __cache__:
        entry = Entry()
        __cache__[request.fs] = entry

    entry = __cache__[request.fs]


    # Process the resource.
    # =====================

    modtime = os.stat(request.fs)[stat.ST_MTIME]
    if entry.modtime == modtime:                            # cache hit
        if entry.exc is not None:
            raise entry.exc
    else:                                                   # cache miss
        try:
            entry.resource = load(request, modtime)
        except:     # capture any Exception
            entry.exc = ( LoadError(traceback.format_exc())
                        , sys.exc_info()[2]
                         )
        else:       # reset any previous Exception
            entry.exc = None 

        entry.modtime = modtime
        if entry.exc is not None:
            raise entry.exc[0] # TODO Why [0] here, and not above?


    # Return
    # ======
    # The caller must take care to avoid mutating any namespace dictionary at 
    # entry.resource.pages[0].

    return entry.resource



class ConfigurationError(StandardError):
    """This is an error in any part of our configuration.
    """

    def __init__(self, msg):
        StandardError.__init__(self)
        self.msg = msg

    def __str__(self):
        return self.msg


class ConfFileError(ConfigurationError):

    def __init__(self, msg, filename, lineno):
        msg = "%s [%s, line %s]" % (msg, lineno, filename)
        ConfigurationError.__init__(self, msg)



from ez_setup import use_setuptools
use_setuptools()

from setuptools import find_packages, setup
version = open('version.txt').read()


classifiers = [ 'Development Status :: 4 - Beta'
              , 'Environment :: Console'
              , 'Intended Audience :: Developers'
              , 'License :: OSI Approved :: MIT License'
              , 'Natural Language :: English'
              , 'Operating System :: MacOS :: MacOS X'
              , 'Operating System :: Microsoft :: Windows'
              , 'Operating System :: POSIX'
              , 'Programming Language :: Python'
              , 'Topic :: Internet :: WWW/HTTP :: HTTP Servers'
               ]

setup( author = 'Chad Whitacre'
     , author_email = 'chad@zetaweb.com'
     , classifiers = classifiers
     , description = ('Aspen is a Python web framework. '
                      'Simplates are the main attraction.')
     , entry_points = { 'console_scripts': [ 'aspen = aspen.server:main'
                                           , 'thrash = aspen:thrash'
                                            ] }
     , name = 'aspen'
     , packages = find_packages(exclude=[ 'aspen.socket_io'     # not ready for
                                        , 'aspen.socket_io.*'      # prime time
                                        , 'aspen.tests'
                                        , 'aspen.tests.*'
                                         ])
     , url = 'http://aspen.io/'
     , version = version
     , zip_safe = False
     , package_data = {'aspen': [ 'www/*'
                                , 'configuration/hooks.conf'
                                , 'configuration/mime.types'
                                 ]}
      )
